# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import pytest

from searchlet.environments import GridEnvironment, GridState
from searchlet.algorithms import IDDFS


def test_iddfs():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    # Solve the problem
    path = IDDFS.solve(env, start, goal)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")


if __name__ == '__main__':
    pytest.main()


# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from abc import ABC, abstractclassmethod
from searchlet import Environment, Action, State, WeightedEnvironment
from typing import Tuple, List, Union, Optional


class SearchAlgorithm(ABC):

    @abstractclassmethod
    def solve(cls, environment: Environment, start: State, goal: Union[List[State], State], *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        raise NotImplementedError("Search algorithms should have a solve method")


class WeightedSearchAlgorithm(ABC):

    @abstractclassmethod
    def solve(cls, environment: WeightedEnvironment, start: State, goal: State, *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        raise NotImplementedError("Search algorithms should have a solve method")


# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from collections import deque
from searchlet import Environment, Action, State
from .SearchAlgorithm import SearchAlgorithm
from typing import Tuple, List, Union, Optional


class BFS(SearchAlgorithm):

    @classmethod
    def solve(cls, environment: Environment, start: State, goal: Union[List[State], State], *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        """Solve a problem using breadth first search
        
        [description]
        
        :param environment: The environment
        :type environment: Environment
        :param start: The starting state
        :type start: State
        :param goal: The goal state
        :type goal: Union[List[State], State]
        :return: The solution, or None if not found
        :rtype: Tuple[List[Action], Union[float, int]]
        """

        # Define the goal test - the reason that we need to 
        # do this is if the goal is a set of goals, and not just
        # a single element. If it is a single element, then we just
        # return current == goal, however a list needs to be checked
        # for each of the goal elements
        if type(goal) == list:
            def goal_test(current: State, goals: List[State]) -> bool:
                for goal in goals:
                    if current == goal:
                        return True
                return False
        else:
            def goal_test(current: State, goal: State) -> bool:
                return current == goal

        # Define the open-list and the parent map
        open_list = deque()
        parent_map = {}
        closed_list = set()

        # Add the start state to the open list
        open_list.append(start)

        while len(open_list) != 0:
            # Pop the current element from the open list
            current_state = open_list.popleft()
            if current_state in closed_list:
                continue
            else:
                closed_list.add(current_state)

            # Check if the current state is the goal
            if goal_test(current_state, goal):
                # Then we know that we've done it!
                # We need to break
                break
                
            # Generate the next set of possible states and add them to
            # the open list
            for possible_action in environment.get_actions(current_state):
                # Generate
                next_state = environment.apply_action(possible_action, current_state)[0]

                if next_state not in parent_map:
                    parent_map[next_state] = (current_state, possible_action)
                # Add to open list
                open_list.append(next_state)

        if goal_test(current_state, goal):
            # We found the goal!
            # Now we need to extract the proper state
            state_actions: List[Tuple[State, Optional[Action]]] = []
            state_actions.append((current_state, None))
            while current_state != start:
                current_state, action = parent_map[current_state]
                state_actions.append((current_state, action))
            return ((list(reversed(state_actions)), len(state_actions)))

        return (None, -1)



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import pytest
import math

from searchlet.environments import GridEnvironment, GridState
from searchlet.algorithms import AStar, Dijkstra, IDAStar


def test_astar_no_heuristic():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    # Solve the problem
    path = AStar.solve(env, start, goal)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")


def test_astar_heuristic():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    def h_fn(x: GridState, y: GridState):
        return 1000 * math.sqrt((x.location[0] - y.location[0]) ** 2 + (x.location[1] - y.location[1]) ** 2)

    # Solve the problem
    path = AStar.solve(env, start, goal, heuristic=h_fn)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")


def test_dijkstra():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    # Solve the problem
    path = Dijkstra.solve(env, start, goal)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")

def test_idastar_heuristic():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    def h_fn(x: GridState, y: GridState):
        return 1000 * math.sqrt((x.location[0] - y.location[0]) ** 2 + (x.location[1] - y.location[1]) ** 2)

    # Solve the problem
    path = IDAStar.solve(env, start, goal, heuristic=h_fn)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")


def test_idastar_no_heuristic():
    # First create and initialize the environment
    env = GridEnvironment(10, 10)
    env.initialize()

    # Get the start and goal
    start = GridState(0, 0)
    goal = GridState(5, 5)

    # Solve the problem
    path = IDAStar.solve(env, start, goal)

    if path[0] is None:
        raise AssertionError("BFS could not find feasible path in solvable environment")

if __name__ == '__main__':
    pytest.main()


if __name__ == '__main__':
    pytest.main()


# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from .Stack import Stack  # noqa: F401
from .PriorityQueue import PriorityQueue  # noqa: F401


# Copyright (c) 2018 David Chan
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from typing import Any


class Stack(object):
    """Simple stack class
    """

    def __init__(self,):
        self.memory = []

    def push(self, obj: Any) -> None:
        """Push an element onto the stack

        :param obj: The object to push onto the stack
        :type obj: Any
        """

        self.memory.append(obj)

    def pop(self,) -> Any:
        """Pops an element from the stack

        :return: The popped element
        :rtype: Any
        """

        ret_val = self.memory[-1]
        self.memory = self.memory[:-1]
        return ret_val

    def peek(self,) -> Any:
        """Peeks at the possible return value of the stack

        :return: The element at the top of the stack
        :rtype: Any
        """

        return self.memory[-1]

    def empty(self, ) -> bool:
        """Returns if the stack is empty
        
        :return: If the stack is empty
        :rtype: bool
        """
        return len(self.memory) == 0

    def __str__(self, ) -> str:
        return str(self.memory)




"""Utilities for exploring envirionments
"""

import random

from searchlet import Environment, State
from typing import List

def generate_random_rollout(env: Environment, state: State, depth: int) -> List[State]:
    """Generate a random rollout from an environment given the starting state and depth
    
    :param env: The environment to use
    :type env: Environment
    :param state: The starting state of the rollout
    :type state: State
    :param depth: The depth of the rollout
    :type depth: int
    :return: The list of states in the rollout
    :rtype: List[State]
    """
    states = []
    states.append(state)
    for _ in range(depth):
        action = random.choice(env.get_actions(state))
        state = env.apply_action(action, state)[0]
        states.append(state)
 
    return states



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from .DFS import DFS, DepthLimitedDFS  # noqa: 401
from .BFS import BFS  # noqa: 401
from .IDDFS import IDDFS  # noqa: 401
from .AStar import AStar  # noqa: 401
from .Dijkstra import Dijkstra  # noqa: 401
from .IDAStar import IDAStar  # noqa: 401



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from searchlet import Action, State, WeightedEnvironment
from .SearchAlgorithm import WeightedSearchAlgorithm
from .AStar import AStar
from typing import Tuple, List, Union, Optional


class Dijkstra(WeightedSearchAlgorithm):

    @classmethod
    def solve(cls, environment: WeightedEnvironment, start: State, goal: Union[List[State], State], *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        """Solve a problem using Dijkstra's algorithm
        
        :param environment: The environment
        :type environment: Environment
        :param start: The starting state
        :type start: State
        :param goal: The goal state
        :type goal: Union[List[State], State]
        :return: The solution, or None if not found
        :rtype: Tuple[List[Action], Union[float, int]]
        """

        return AStar.solve(environment, start, goal, heuristic=None, *args, **kwargs)


# Copyright (c) 2018 David Chan
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from enum import Enum
from abc import ABC, abstractclassmethod
from typing import List, Tuple, Optional, Union


class Action(object):
    """
    Abstract base class representing an action type. Primarily
    for typing, as actions are traditionally handled on an
    environment by environment basis.
    """

    def __init__(self, *args, **kwargs) -> None:
        pass


class ReverseAction(Action):
    """
    Abstract base class representing a reverse action. In general, the
    reverse action is an action which when applied, undoes an action
    which was generated earlier. This is one of the faster ways
    to handle traversing a search space
    """

    def __init__(self, *args, **kwargs) -> None:
        pass


class State(ABC):
    """
    Abstract base class representing an environment state. The environment
    is separated from the state, as it makes it easier to abstract the search
    algorithms.
    """

    def __init__(self, ) -> None:
        pass


class Environment(ABC):
    """
    Abstract base class representing a search environment. This
    is a key part of a search problem, which we can solve using
    other methods in the library
    """

    def __init__(self, ) -> None:
        pass

    @abstractclassmethod
    def initialize(self, *args, **kwargs) -> None:
        raise NotImplementedError("Environments should have an initialization method, which sets the value of the start.")

    @abstractclassmethod
    def get_actions(self, state: State) -> List[Action]:
        raise NotImplementedError("Environments must have a get_actions method.")

    @abstractclassmethod
    def apply_action(self, action: Action, state: State) -> Tuple[State, Optional[Action]]:
        raise NotImplementedError("Environments must have an apply_action method.")


class WeightedEnvironment(Environment):
    """
    Abstract class representing a weighted environment. In these environments
    we have additional edge-cost parameters, which are not necessarily present
    in the traditional search environment.
    """

    def __init__(self, ) -> None:
        pass

    @abstractclassmethod
    def get_value(self, state: State) -> Union[float, int]:
        raise NotImplementedError("No unary value method defined in a weighted edge-cost environment")

    @abstractclassmethod
    def get_cost(self, action: Action, state: State) -> Union[float, int]:
        raise NotImplementedError("No action value method defined in a weighted edge-cost environment.")



# Copyright (c) 2018 David Chan
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import queue
import math

from .SearchAlgorithm import WeightedSearchAlgorithm
from searchlet import WeightedEnvironment, Action, State
from searchlet.ds import PriorityQueue
from typing import Union, List, Tuple, Optional, Callable


class IDAStar(WeightedSearchAlgorithm):

    @classmethod
    def _bounded_astar_solve(cls,
                             environment: WeightedEnvironment,
                             start: State,
                             goal: State,
                             heuristic: Callable[[State, State], Union[float, int]],
                             f_cost_bound: Union[float, int]) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:

        # Initialize some variables
        open_list = PriorityQueue()
        closed_list = {}
        parent_map = {}

        # Add the start state to the queue
        open_list.push((start, 0), 0 + heuristic(start, goal))

        min_over_f_cost = math.inf

        # Main Loop
        while not open_list.empty():

            # Pop the lowest f-cost element from the queue
            f_cost, (current_state, g_cost) = open_list.pop()

            if f_cost > f_cost_bound:
                min_over_f_cost = min(f_cost, min_over_f_cost)
                continue

            # Check to see if this is the goal
            if current_state == goal:
                # We're done!
                break

            # If this is not the goal, expand the neighbor nodes
            for action in environment.get_actions(current_state):
                # Get the next state
                next_state = environment.apply_action(action, current_state)[0]

                # Get the cost of the next state
                cost = environment.get_cost(action, current_state)

                # Compute the f_cost of the new state
                new_f_cost = g_cost + cost + heuristic(next_state, goal)

                if next_state not in closed_list or closed_list[next_state] > new_f_cost:
                    # Add it to the priority queue and parent map
                    # the priority queue implementation will not update unless it has to
                    return_value = open_list.push(
                        (next_state, g_cost + cost), priority=new_f_cost)
                    if return_value == PriorityQueue.NONEXIST or return_value == PriorityQueue.EXISTS_UPDATED:
                        # Only add to the parent map if our way is better
                        parent_map[next_state] = (current_state, action, cost)

            # Add the element to the closed list
            closed_list[current_state] = f_cost

        if current_state == goal:
            # We solved the problem!
            total_cost = 0
            state_action_pairs = []
            state_action_pairs.append((current_state, None))
            while current_state != start:
                current_state, action, cost = parent_map[current_state]
                total_cost += cost
                state_action_pairs.append((current_state, action))

            return (list(reversed(state_action_pairs)), total_cost)

        return (None, min_over_f_cost)

    @classmethod
    def solve(cls,
              environment: WeightedEnvironment,
              start: State,
              goal: State,
              *args,
              **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:

        # Handle the heuristic function
        if 'heuristic' in kwargs and kwargs['heuristic'] is not None:
                h_function = kwargs['heuristic']
        else:
            def h_function(x: State, y: State) -> Union[float, int]:
                return 0

        # Main solution loop
        heuristic_start_bound = h_function(start, goal)
        while True:
            path = IDAStar._bounded_astar_solve(
                environment, start, goal, h_function, heuristic_start_bound)
            if path[0] is None:
                heuristic_start_bound = path[1]
            else:
                return path



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from searchlet import Environment, Action, State
from .SearchAlgorithm import SearchAlgorithm
from .DFS import DepthLimitedDFS
from typing import Tuple, List, Union, Optional
import math


class IDDFS(SearchAlgorithm):

    @classmethod
    def solve(cls, environment: Environment, start: State, goal: Union[List[State], State], *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        """Solve a problem using iterative-deepening depth-limited depth first search
        
        :param environment: The environment
        :type environment: Environment
        :param start: The starting state
        :type start: State
        :param goal: The goal state
        :type goal: Union[List[State], State]
        :param max_depth: The maximum depth of the IDDFS search
        :type max_depth: int
        :return: The solution, or None if not found
        :rtype: Tuple[List[Action], Union[float, int]]
        """

        # Handle the maximum depth parameter
        if 'max_depth' in kwargs:
            max_depth = kwargs['max_depth']
        else:
            max_depth = math.inf

        depth = 0
        while depth < max_depth:
            depth += 1
            path = DepthLimitedDFS.solve(environment, start, goal, depth_limit=depth, *args, **kwargs)
            if path[0] is None and path[1] == -1:
                return path
            elif path[0] is not None:
                return path
class IDDFS(SearchAlgorithm):

    @classmethod
    def solve(cls, environment: Environment, start: State, goal: Union[List[State], State], *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        """Solve a problem using iterative-deepening depth-limited depth first search
        
        :param environment: The environment
        :type environment: Environment
        :param start: The starting state
        :type start: State
        :param goal: The goal state
        :type goal: Union[List[State], State]
        :param max_depth: The maximum depth of the IDDFS search
        :type max_depth: int
        :return: The solution, or None if not found
        :rtype: Tuple[List[Action], Union[float, int]]
        """

        # Handle the maximum depth parameter
        if 'max_depth' in kwargs:
            max_depth = kwargs['max_depth']
        else:
            max_depth = math.inf

        depth = 0
        while depth < max_depth:
            depth += 1
            path = DepthLimitedDFS.solve(environment, start, goal, depth_limit=depth, *args, **kwargs)
            if path[0] is None and path[1] == -1:
                return path
            elif path[0] is not None:
                return path


# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import itertools
import heapq

from typing import List, Any, Union


class PriorityQueue(object):

    REMOVED = '<removed-element>'
    EXISTS_LOWER_PRIORITY = 1
    EXISTS_UPDATED = 2
    NONEXIST = 0

    def __init__(self, ) -> None:
        self.memory: List[Any] = []
        self.counter = itertools.count()
        self.size = 0
        self.map = {}

    def push(self, element: Any, priority: Union[float, int]=0) -> int:
        return_value = PriorityQueue.NONEXIST
        if element in self.map:
            if self.map[element][0] < priority:
                return PriorityQueue.EXISTS_LOWER_PRIORITY
            self.remove_element(element)
            return_value = PriorityQueue.EXISTS_UPDATED
        else:
            self.size += 1
        count = next(self.counter)
        entry = [priority, count, element]
        self.map[element] = entry
        heapq.heappush(self.memory, entry)
        return return_value

    def remove_element(self, element) -> None:
        entry = self.map.pop(element)
        entry[-1] = PriorityQueue.REMOVED
    
    def pop(self, ) -> Any:
        while self.memory:
            priority, _, element = heapq.heappop(self.memory)
            if element is not PriorityQueue.REMOVED:
                del self.map[element]
                self.size -= 1
                return (priority, element)
        raise KeyError("Tried to pop from an empty queue")
    
    def empty(self, ) -> bool:
        return self.size <= 0



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import pytest
import os

from .Environment import Action, ReverseAction, State, Environment, WeightedEnvironment  # noqa: F401

def test():
    print(os.path.join(*(__file__.split('__init__.py'))))
    pytest.main([os.path.join(*(__file__.split('__init__.py')))])


# Copyright (c) 2018 David Chan
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from typing import List, Tuple, Optional, cast
from enum import Enum
from searchlet import Action, WeightedEnvironment, State


class GridAction(Action, Enum):
    UP = 1
    DOWN = 2
    LEFT = 3
    RIGHT = 4


class GridState(State):
    def __init__(self, loc_x: int, loc_y: int) -> None:
        self.location = (loc_x, loc_y)
    
    def __eq__(self, other):
        return self.location[0] == other.location[0] and self.location[1] == other.location[1]
    
    def __hash__(self,):
        return self.location.__hash__()

    def __str__(self,):
        return str(self.location)

    def __repr__(self, ):
        return repr(self.location)


class GridEnvironment(WeightedEnvironment):

    def __init__(self, height: int, width: int) -> None:
        self.height = height
        self.width = width

    def initialize(self, *args, **kwargs) -> None:
        pass

    def get_actions(self, state: State) -> List[Action]:
        state = cast(GridState, state)
        actions: List[Action] = []
        if state.location[1] > 0:
            actions.append(GridAction.DOWN)
        if state.location[1] < self.height:
            actions.append(GridAction.UP)
        if state.location[0] > 0:
            actions.append(GridAction.LEFT)
        if state.location[0] < self.width:
            actions.append(GridAction.RIGHT)

        return actions

    def apply_action(self, action: Action, state: State) -> Tuple[State, Optional[Action]]:
        inv_action = None
        state = cast(GridState, state)
        x_delta = 0
        y_delta = 0
        if action is GridAction.UP:
            y_delta += 1
            inv_action = GridAction.DOWN
        elif action is GridAction.DOWN:
            y_delta -= 1
            inv_action = GridAction.UP
        elif action is GridAction.LEFT:
            x_delta -= 1
            inv_action = GridAction.RIGHT
        elif action is GridAction.RIGHT:
            x_delta += 1
            inv_action = GridAction.LEFT

        new_state = GridState(state.location[0] + x_delta, state.location[1] + y_delta)

        return (new_state, inv_action)

    def get_value(self, state: State) -> float:
        return 0

    def get_cost(self, action: Action, state: State) -> float:
        return 1



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='searchlet',
    version='1.0.0',
    description='Python Search Algorithms',
    long_description=readme,
    author='David Chan',
    author_email='davidchan@berkeley.edu',
    url='https://github.com/DavidMChan/searchlet',
    license=license,
    install_requires=[
        'pytest >= 3.7.0',
    ],
    packages=find_packages(),  # exclude=('tests', 'docs')
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# Copyright (c) 2018 David Chan
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import queue

from .SearchAlgorithm import WeightedSearchAlgorithm
from searchlet import WeightedEnvironment, Action, State
from searchlet.ds import PriorityQueue
from typing import Union, List, Tuple, Optional

class AStar(WeightedSearchAlgorithm):

    @classmethod
    def solve(cls, environment: WeightedEnvironment, start: State, goal: State, *args, **kwargs) -> Tuple[Optional[List[Tuple[State, Optional[Action]]]], Union[float, int]]:
        """Solve a problem using the A* algorithm

        TODO: Support more than one goal state
        
        :param environment: The environment
        :type environment: Environment
        :param start: The starting state
        :type start: State
        :param goal: The goal state
        :type goal: Union[List[State], State]
        :param heuristic: The heuristic function to use
        :type heuristic: Callable[[State, State], Union[float, int]]
        :return: The solution, or None if not found
        :rtype: Tuple[List[Action], Union[float, int]]
        """

        # Handle the heuristic function
        if 'heuristic' in kwargs and kwargs['heuristic'] is not None:
                h_function = kwargs['heuristic']
        else:
            def h_function(x: State, y: State) -> Union[float, int]:
                return 0
        
        # Initialize some variables
        open_list = PriorityQueue()
        closed_list = {}
        parent_map = {}

        # Add the start state to the queue
        open_list.push((start, 0), 0 + h_function(start, goal))

        # Main Loop
        while not open_list.empty():

            # Pop the lowest f-cost element from the queue
            f_cost, (current_state, g_cost) = open_list.pop()

            # Check to see if this is the goal
            if current_state == goal:
                # We're done!
                break
            
            # If this is not the goal, expand the neighbor nodes
            for action in environment.get_actions(current_state):
                # Get the next state
                next_state = environment.apply_action(action, current_state)[0]
                
                # Get the cost of the next state
                cost = environment.get_cost(action, current_state)

                # Compute the f_cost of the new state
                new_f_cost = g_cost + cost + h_function(next_state, goal)

                if next_state not in closed_list or closed_list[next_state] > new_f_cost:
                    # Add it to the priority queue and parent map
                    # the priority queue implementation will not update unless it has to
                    return_value = open_list.push((next_state, g_cost + cost), priority=new_f_cost)
                    if return_value == PriorityQueue.NONEXIST or return_value == PriorityQueue.EXISTS_UPDATED:
                        # Only add to the parent map if our way is better
                        parent_map[next_state] = (current_state, action, cost)

            # Add the element to the closed list
            closed_list[current_state] = f_cost

        if current_state == goal:
            # We solved the problem!
            total_cost = 0
            state_action_pairs = []
            state_action_pairs.append((current_state, None))
            while current_state != start:
                current_state, action, cost = parent_map[current_state]
                total_cost += cost
                state_action_pairs.append((current_state, action))
            
            return (list(reversed(state_action_pairs)), total_cost)
        
        return (None, -1)
